from lazyqml.GUI._widgets import *

import ipywidgets

# Groups options in boxes
general_model_options_box = ipywidgets.VBox(
    [
        ipywidgets.Label("General model options"),
        epochs_widget,
        layers_widget,
        nqubits_widget,
        dataset_widget
    ], 
    layout=ipywidgets.Layout(width='50%', margin='10pt')
)

cv_options_box = ipywidgets.VBox(
    [
        ipywidgets.Label("Cross-validation options"),
        cv_checkbox,
        splits_widget,
        repeats_widget
    ],
    layout=ipywidgets.Layout(width='50%', margin='10pt')
)

classifiers_box = ipywidgets.VBox(
    [
        ipywidgets.Label("Classifiers"),
        qsvm_checkbox,
        qnn_checkbox,
        qnn_bag_checkbox
    ], 
    layout=ipywidgets.Layout(width='50%', margin='10pt')
)

ansatzs_box = ipywidgets.VBox(
    [
        ipywidgets.Label("Ansatzs"),
        hp_checkbox,
        tt_checkbox,
        two_checkbox,
        hwe_checkbox,
        annular_checkbox
    ],
    layout=ipywidgets.Layout(width='50%', margin='10pt')
)

embeddings_box = ipywidgets.VBox(
    [
        ipywidgets.Label("Embeddings"),
        rx_checkbox,
        ry_checkbox,
        rz_checkbox,
        zz_checkbox,
        amp_checkbox,
        dense_checkbox,
        ho_checkbox
    ],
    layout=ipywidgets.Layout(width='50%', margin='10pt')
)

bagging_box = ipywidgets.VBox(
    [
        ipywidgets.Label("Bagging options"),
        nestimators_widget,
        nsamples_widget,
        nfeatures_widget
    ],
    layout=ipywidgets.Layout(width='100%', margin='10pt')
)

other_options_box = ipywidgets.VBox(
    [
        ipywidgets.Label("Other options"),
        tn_widget,
        runs_widget,
        randomstate_widget,
        verbose_widget
    ],
    layout=ipywidgets.Layout(width='100%', margin='10pt')
)


"""Top-level package for lazyqml."""

__author__ = """Diego García Vega, Fernando Álvaro Plou Llorente, Alejandro Leal Castaño"""
__email__ = "garciavdiego@uniovi.es, ploufernando@uniovi.es, lealcalejandro@uniovi.es"
__version__ = "0.0.17"

from .lazyqml import QuantumClassifier

import torch
torch.set_num_threads(1)
torch.set_num_interop_threads(1)

__all__ = ['QuantumClassifier']


import warnings

from lazyqml.Factories import CircuitFactory
from lazyqml.Global.globalEnums import Backend
from lazyqml.Utils.Utils import get_max_bond_dim, get_simulation_type

import numpy     as np
import pennylane as qml
import torch
import torch.nn as nn
import torch.utils.data

class QNNTorch:
    def __init__(
        self,
        nqubits,
        ansatz,
        embedding,
        n_class,
        layers,
        epochs,
        shots,
        lr,
        batch_size   = 10,
        torch_device = "cpu",
        backend      = "lightning.qubit",
        diff_method  = "best",
        seed         = 1234,
    ):
        torch.manual_seed(seed)
        np.random.seed(seed)

        self.nqubits      = nqubits
        self.ansatz       = ansatz
        self.embedding    = embedding
        self.n_class      = n_class
        self.layers       = layers
        self.epochs       = epochs
        self.lr           = lr
        self.batch_size   = batch_size
        self.torch_device = torch_device
        self.backend      = backend
        self.diff_method  = diff_method
        self.shots        = shots
        self.circuit_factory = CircuitFactory(nqubits, layers)

        warnings.filterwarnings("ignore")

        self._build_device()
        self._build_qnode()

        self.criterion = nn.BCEWithLogitsLoss() if n_class == 2 else nn.CrossEntropyLoss()
        self.params    = None
        self.opt       = None

    @property
    def n_params(self):
        return self._n_params

    def _build_device(self):
        # Create device
        if get_simulation_type() == "tensor":
            if self.backend != Backend.lightningTensor:
                device_kwargs = {
                    "max_bond_dim": get_max_bond_dim(),
                    "cutoff": np.finfo(np.complex128).eps,
                    # "contract": "auto-mps",
                }
            else:
                device_kwargs = {
                    "max_bond_dim": get_max_bond_dim(),
                    "cutoff": 1e-10,
                    "cutoff_mode": "abs",
                }
            
            self.dev = qml.device(self.backend, wires=self.nqubits, method='mps', **device_kwargs)
        else:
            self.dev = qml.device(self.backend, wires=self.nqubits)

    def _build_qnode(self):
        wires = range(self.nqubits)

        ansatz = self.circuit_factory.GetAnsatzCircuit(self.ansatz)
        embedding = self.circuit_factory.GetEmbeddingCircuit(self.embedding)

        def circuit(x, params):
            # self.embedding(x, wires=wires)
            # self.ansatz(params, wires=wires, nlayers=self.layers)

            embedding(x, wires=wires)
            ansatz.getCircuit()(params, wires=wires)

            if self.n_class == 2:
                return qml.expval(qml.PauliZ(0))
            return tuple(qml.expval(qml.PauliZ(i)) for i in range(self.n_class))

        # QNode base (sin batch)
        base_qnode = qml.QNode(circuit, self.dev, interface="torch", diff_method=self.diff_method)

        # Batching portable
        self.qnode = qml.batch_input(base_qnode, argnum=0)

        # Retrieve number of parameters from the ansatz
        self._n_params = ansatz.n_total_params

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        y = self.qnode(x, self.params)

        if self.n_class == 2:
            return y.reshape(-1, 1)

        if isinstance(y, (tuple, list)):
            y = torch.stack(list(y), dim=0)  # (n_class, batch)

        if y.ndim == 2 and y.shape[0] == self.n_class:
            y = y.transpose(0, 1)  # (batch, n_class)
        return y

    def fit(self, X, y):
        X_train = torch.tensor(X, dtype=torch.float32).to(self.torch_device)
        y_train = torch.tensor(y, dtype=torch.float32 if self.n_class == 2 else torch.long).to(self.torch_device)

        if self.n_class == 2 and y_train.ndim == 1:
            y_train = y_train.unsqueeze(1)

        self.params = torch.randn((self.n_params,), device=self.torch_device, requires_grad=True)
        self.opt    = torch.optim.Adam([self.params], lr=self.lr)

        ds          = torch.utils.data.TensorDataset(X_train, y_train)
        data_loader = torch.utils.data.DataLoader(ds, batch_size=self.batch_size, shuffle=True, drop_last=True)

        for _epoch in range(self.epochs):
            for batch_X, batch_y in data_loader:
                self.opt.zero_grad(set_to_none=True)
                preds = self.forward(batch_X)
                loss  = self.criterion(preds, batch_y)
                loss.backward()
                self.opt.step()
        self.params = self.params.detach()

    def predict(self, X):
        X_test = torch.tensor(X, dtype=torch.float32).to(self.torch_device)

        preds_all = []
        bs        = max(1, self.batch_size)
        with torch.inference_mode():
            for i in range(0, X_test.shape[0], bs):
                preds_all.append(self.forward(X_test[i:i + bs]))
        y_pred = torch.cat(preds_all, dim=0)

        if self.n_class == 2:
            y_pred = torch.sigmoid(y_pred.view(-1))
            return (y_pred > 0.5).cpu().numpy()

        return torch.argmax(y_pred, dim=1).cpu().numpy()


from lazyqml.Interfaces.iAnsatz import Ansatz
import pennylane as qml

class HardwareEfficient(Ansatz):
    def getCircuit(self):
        def hardware_efficient_ansatz(theta, wires):
            """Implements a hardware-efficient ansatz circuit.

            Args:
                theta (array[float]): array of parameters for the ansatz circuit
                wires (Sequence[int]): wires that the ansatz circuit acts on

            Returns:
                None
            """
            param_count = 0
            
            for _ in range(self.nlayers):

                N = len(wires)
                assert len(theta) == 3 * N * self.nlayers
                
                for i in range(N):
                    qml.RX(theta[param_count], wires=wires[i])
                    param_count +=1
                
                for i in range(N-1):
                    qml.CNOT(wires=[wires[i], wires[i + 1]])
                qml.CNOT(wires=[wires[N-1],wires[0]])
                
                for i in range(N):
                    qml.RZ(theta[param_count], wires=wires[i])
                    param_count +=1
                
                for i in range(N-1):
                    qml.CNOT(wires=[wires[i], wires[i + 1]])
                qml.CNOT(wires=[wires[N-1],wires[0]])

                for i in range(N):
                    qml.RX(theta[param_count], wires=wires[i])
                    param_count +=1
                
                for i in range(N-1):
                    qml.CNOT(wires=[wires[i], wires[i + 1]])
                qml.CNOT(wires=[wires[N-1],wires[0]])

        return hardware_efficient_ansatz
    
    @property
    def n_ansatz_params(self):
        return 3 * self.nqubits


# Imports
import pandas as pd
import numpy as np
import torch
import psutil
import GPUtil
from itertools import product
from sklearn.model_selection import LeaveOneOut, StratifiedKFold, train_test_split
from threadpoolctl           import threadpool_limits, threadpool_info

# Importing from
from lazyqml.Global.globalEnums import *
import lazyqml.Global._config as cfg

"""
------------------------------------------------------------------------------------------------------------------
    Verbose printer class
        - This class implements the functionlity to print or not depending on a boolean flag
        - The message is preceded by "[VERBOSE] {message}"
        - It is implemented as a Singleton Object
------------------------------------------------------------------------------------------------------------------
"""
class VerbosePrinter:
    _instance = None
    _initialized = False

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(VerbosePrinter, cls).__new__(cls)
        return cls._instance

    def __init__(self):
        if not VerbosePrinter._initialized:
            self.verbose = False
            VerbosePrinter._initialized = True

    def set_verbose(self, verbose: bool):
        self.verbose = verbose

    def print(self, message: str):
        if self.verbose:
            print(f"[VERBOSE] {message}")
        else:
            pass

    @classmethod
    def get_instance(cls):
        if cls._instance is None:
            cls._instance = VerbosePrinter()
        return cls._instance

"""
------------------------------------------------------------------------------------------------------------------
                                            Miscelaneous Utils
------------------------------------------------------------------------------------------------------------------
"""

def adjustQubits(nqubits, numClasses):
    """
        Adjust the number of qubits to be able to solve the problem
    """
    # Find the next power of 2 greater than numClasses
    power = np.ceil(np.log2(numClasses))
    nqubits = 2 ** power
    # Ensure nqubits is greater than numClasses
    if nqubits <= numClasses:
        nqubits *= 2
    return int(nqubits)

def calculate_quantum_memory(num_qubits, overhead=2):
    """
        Estimates the memory in MiB used by the quantum circuits.
    """
    # Each qubit state requires 2 complex numbers (amplitude and phase)
    # Each complex number uses 2 double-precision floats (16 bytes)
    bytes_per_qubit_state = 16

    # Number of possible states is 2^n, where n is the number of qubits
    if get_simulation_type() == "statevector":
        num_states = 2 ** num_qubits
    else:
        num_states =  num_qubits * (get_max_bond_dim() ** 2)

    # Total memory in bytes
    total_memory_bytes = num_states * bytes_per_qubit_state * overhead

    # Convert to more readable units

    return total_memory_bytes / (1024**2)

def calculate_free_memory():
    """
        Calculates the amount of free RAM
    """
    # Use psutil to get available system memory (in MiB)
    mem = psutil.virtual_memory()
    free_ram_mb = mem.available / (1024 ** 2)  # Convert bytes to MiB
    return free_ram_mb

def calculate_free_video_memory():
    """
    Calculates the amount of free Video Memory.
    """
    try:
        gpus = GPUtil.getGPUs()
        if not gpus:
            raise ValueError("No GPUs found.")
        return gpus[0].memoryFree
    except Exception as e:
        print(f"Error calculating free video memory: {e}")
        return 0  # Return None or an appropriate default value

def create_combinations(classifiers, embeddings, ansatzs, features, qubits, folds, repeats):
    classifier_list = []
    embedding_list = []
    ansatzs_list = []

    # Make sure we don't have duplicated items
    classifiers = list(classifiers)
    embeddings = list(embeddings)
    ansatzs = list(ansatzs)
    qubit_values = sorted(list(qubits))
    repeat_range = range(repeats)
    folds_range = range(folds)
    
    cv_size = repeats*folds

    if Model.ALL in classifiers:
        classifier_list = Model.list()
        classifier_list.remove(Model.ALL)
    else:
        classifier_list = classifiers

    if Embedding.ALL in embeddings:
        embedding_list = Embedding.list()
        embedding_list.remove(Embedding.ALL)
    else:
        embedding_list = embeddings

    if Ansatzs.ALL in ansatzs:
        ansatzs_list = Ansatzs.list()
        ansatzs_list.remove(Ansatzs.ALL)
    else:
        ansatzs_list = ansatzs

    combo_counter = 0
    combinations = []
    # Create all base combinations first
    for qubits in qubit_values:
        for classifier in classifier_list:
            temp_combinations = []
            if classifier == Model.QSVM or classifier == Model.QKNN:
                # QSVM doesn't use ansatzs or features but uses qubits
                temp_combinations = list(product([qubits], [classifier], embedding_list, [None], [None], repeat_range, folds_range))
            elif classifier == Model.QNN:
                # QNN uses ansatzs and qubits
                temp_combinations = list(product([qubits], [classifier], embedding_list, ansatzs_list, [None], repeat_range, folds_range))
            elif classifier == Model.QNN_BAG:
                # QNN_BAG uses ansatzs, features, and qubits
                temp_combinations = list(product([qubits], [classifier], embedding_list, ansatzs_list, features, repeat_range, folds_range))

            # Add memory calculation for each combination
            for combo in temp_combinations:
                memory = calculate_quantum_memory(combo[0])  # Calculate memory based on number of qubits
                combinations.append((combo_counter // cv_size, *combo, memory))
                combo_counter += 1

    return combinations

def fixSeed(seed):
    np.random.seed(seed=seed)
    torch.manual_seed(seed)

def generate_cv_indices(X, y, mode="cross-validation", test_size=0.4, n_splits=5, n_repeats=1, random_state=None):
    """
    Generate train and test indices for either cross-validation, holdout split, or leave-one-out.

    Parameters:
        X (pd.DataFrame or np.ndarray): The features matrix.
        y (pd.Series or np.ndarray): The target vector.
        mode (str): "cross-validation", "holdout", or "leave-one-out".
        test_size (float): Test set proportion for the holdout split (ignored for CV and LOO).
        n_splits (int): Number of folds in StratifiedKFold (ignored for holdout and LOO).
        n_repeats (int): Number of repeats for cross-validation (ignored for holdout and LOO).
        random_state (int): Random state for reproducibility.

    Returns:
        dict: A dictionary of train/test indices.
    """
    cv_indices = {}

    if mode == "hold-out":
        # Single train-test split for holdout
        train_idx, test_idx = train_test_split(
            np.arange(len(X)),
            test_size=test_size,
            stratify=y,
            random_state=random_state
        )
        cv_indices[(0, 0)] = {
            'train_idx': train_idx,
            'test_idx': test_idx
        }

    elif mode == "cross-validation":
        # StratifiedKFold for cross-validation splits
        for repeat in range(n_repeats):
            skf = StratifiedKFold(n_splits=n_splits, shuffle=True, random_state=random_state + repeat if random_state is not None else None)
            for fold, (train_idx, test_idx) in enumerate(skf.split(X, y)):
                cv_indices[(repeat, fold)] = {
                    'train_idx': train_idx,
                    'test_idx': test_idx
                }

    elif mode == "leave-one-out":
        # LeaveOneOut cross-validation
        loo = LeaveOneOut()
        for fold, (train_idx, test_idx) in enumerate(loo.split(X)):
            cv_indices[(0, fold)] = {
                'train_idx': train_idx,
                'test_idx': test_idx
            }

    else:
        raise ValueError("Invalid mode. Choose 'hold-out', 'cross-validation', or 'leave-one-out'.")

    return cv_indices

def get_train_test_split(cv_indices, repeat_id=0, fold_id=0):
    """
    Retrieve the train and test indices for a given repeat and fold ID.

    Parameters:
        cv_indices (dict): The cross-validation indices dictionary.
        repeat_id (int): The repeat ID (0 to n_repeats-1 or 0 for holdout/LOO).
        fold_id (int): The fold ID within the specified repeat.

    Returns:
        tuple: (train_idx, test_idx) arrays for the specified fold and repeat.
    """
    indices = cv_indices.get((repeat_id, fold_id))
    if indices is None:
        print(f"RepeatID {repeat_id}, FoldID{fold_id}")
        raise ValueError("Invalid repeat_id or fold_id specified.")

    return indices['train_idx'], indices['test_idx']

def dataProcessing(X, y, prepFactory, customImputerCat, customImputerNum,
                train_idx, test_idx, ansatz=None, embedding=None):
    """
    Process data for specific train/test indices.

    Parameters:
    - X: Input features
    - y: Target variable
    - prepFactory: Preprocessing factory object
    - customImputerCat: Categorical imputer
    - customImputerNum: Numerical imputer
    - train_idx: Training set indices
    - test_idx: Test set indices
    - ansatz: Optional preprocessing ansatz
    - embedding: Optional embedding method

    Returns:
    Tuple of (X_train_processed, X_test_processed, y_train, y_test)
    """
    # Split the data using provided indices
    X_train, X_test = X.iloc[train_idx], X.iloc[test_idx]
    y_train, y_test = y[train_idx], y[test_idx]

    # Create sanitizer and preprocess
    sanitizer = prepFactory.GetSanitizer(customImputerCat, customImputerNum)
    X_train = pd.DataFrame(sanitizer.fit_transform(X_train))
    X_test = pd.DataFrame(sanitizer.transform(X_test))

    # Apply additional preprocessing if ansatz/embedding provided
    if ansatz is not None or embedding is not None:
        preprocessing = prepFactory.GetPreprocessing(ansatz=ansatz, embedding=embedding)
        X_train_processed = np.array(preprocessing.fit_transform(X_train, y=y_train))
        X_test_processed = np.array(preprocessing.transform(X_test))
    else:
        X_train_processed = np.array(X_train)
        X_test_processed = np.array(X_test)

    # Convert target variables to numpy arrays
    y_train = np.array(y_train)
    y_test = np.array(y_test)

    return X_train_processed, X_test_processed, y_train, y_test

######
def get_embedding_expressivity(nqubits, embedding):
    if embedding == Embedding.AMP:
        return 2**nqubits
    elif embedding == Embedding.DENSE_ANGLE:
        return 2*nqubits
    else:
        return nqubits
    
def find_output_shape(model, sample):
    sample = torch.Tensor(sample)

    output = torch.flatten(model(sample))
    return output.shape[0]


######
def _numpy_math_api():
    runtime_cfg = threadpool_info()
    for cfg in runtime_cfg:
        filepath = cfg.get("filepath", "")
        if "numpy" in filepath:
            return cfg.get("user_api", None)
    return None


######
def set_max_bond_dim(dim: int):
    """
    Sets the maximum bond dimension for tensor network simulation.

    Parameters
    ----------
    dim : int
        Maximum bond dimension
    """
    cfg._max_bond_dim = dim

def get_max_bond_dim():
    return cfg._max_bond_dim

def set_simulation_type(sim):
    """
    Sets the qubit representation for quantum circuit simulation.

    Parameters
    ----------
    sim : str
        String that represents the type of simulation. 'tensor' for tensor network simulation and 'statevector' for state vector simulation.
    """
    try:
        assert sim == "statevector" or sim == "tensor"
        cfg._simulation = sim

    except Exception as e:
        raise ValueError(f"Simulation type must be \"statevector\" or \"tensor\". Got \"{sim}\"")
    

def get_simulation_type():
    return cfg._simulation
######

printer = VerbosePrinter()



# from .CustomEmbedding import ZZEmbedding, _DenseAngleEmbedding

from .ZZ import ZZEmbedding
from .DenseAngle import DenseAngleEmbedding
from .HigherOrder import HigherOrderEmbedding

__all__ = ['ZZEmbedding', 'DenseAngleEmbedding', 'HigherOrderEmbedding']


from abc import abstractmethod
from lazyqml.Interfaces.iCircuit import Circuit

class Ansatz(Circuit):
    def __init__(self, nqubits, nlayers):
        self.nqubits = nqubits
        self.nlayers = nlayers

        self._n_ansatz_params = None

    @property
    @abstractmethod
    def n_ansatz_params(self):
        pass
    
    @property
    def n_total_params(self):
        return self.n_ansatz_params * self.nlayers


"""
------------------------------------------------------------------------------------------------------------------
    This file is devoted to define the global enums for easening the interface.
------------------------------------------------------------------------------------------------------------------
"""

from enum import Enum

class BaseEnum(Enum):
    @classmethod
    def list(cls):
        return list(cls)

# Enum for selecting the Ansatz circuits
class Ansatzs(BaseEnum):
    ALL = 1
    HCZRX = 2
    TREE_TENSOR = 3
    TWO_LOCAL = 4
    HARDWARE_EFFICIENT = 5
    ANNULAR = 6

# Enum for selecting the Embedding circuits
class Embedding(BaseEnum):
    ALL = 1
    RX = 2
    RZ = 3
    RY = 4
    ZZ = 5
    AMP = 6
    DENSE_ANGLE = 7
    HIGHER_ORDER = 8

# Enum for selecting the Models
class Model(BaseEnum):
    ALL = 1
    QNN = 2
    QNN_BAG = 3
    QSVM = 4
    QKNN = 5

class Backend(Enum):
    defaultQubit = "default.qubit"
    
    lightningQubit = "lightning.qubit"
    lightningGPU = "lightning.gpu"

    defaultTensor = "default.tensor"
    lightningTensor = "lightning.tensor"


from .HardwareEfficient import HardwareEfficient
from .HCzRx import HCzRx
from .TreeTensor import TreeTensor
from .TwoLocal import TwoLocal
from .Annular import Annular

__all__ = ['HardwareEfficient', 'HCzRx', 'TreeTensor', 'TwoLocal', 'Annular']


# Importing from
from lazyqml.Interfaces.iPreprocessing import Preprocessing

from sklearn.compose import ColumnTransformer
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.compose import make_column_selector as selector
from sklearn.preprocessing   import MinMaxScaler, RobustScaler

import numpy as np

class Sanitizer(Preprocessing):
    def __init__(self, imputerCat, imputerNum):
        # scalers = [("robust", RobustScaler()),
        #            ("scaler", MinMaxScaler(feature_range=(0, 2*np.pi), clip=True))]
        
        scalers = [("scaler", StandardScaler())]
        
        cat_steps = [("imputer", imputerCat)] + scalers
        num_steps = [("imputer", imputerNum)] + scalers

        self.categorical_transformer = Pipeline(steps=cat_steps)
        self.numeric_transformer = Pipeline(steps=num_steps)

        self.preprocessor = ColumnTransformer(
            transformers=[
                ("numeric", self.numeric_transformer, selector(dtype_exclude="category")),
                ("categorical_low", self.categorical_transformer, selector(dtype_include="category")),
            ]
        )

    def fit(self, X):
        return self.preprocessor.fit(X)

    def fit_transform(self, X):
        return self.preprocessor.fit_transform(X)

    def transform(self, X):
        return self.preprocessor.transform(X)


from abc import ABC, abstractmethod

class Preprocessing(ABC):
    @abstractmethod
    def fit(X, y, self):
        pass
    
    @abstractmethod
    def fit_transform(X, y, self):
        pass

    @abstractmethod
    def transform(X, self):
        pass


from lazyqml.Interfaces.iModel import Model
import numpy as np
from sklearn.svm import SVC
import pennylane as qml
from lazyqml.Factories import CircuitFactory
from lazyqml.Utils import printer

from functools import partial

from time import time

class QSVM(Model):
    def __init__(self, nqubits, embedding, backend, shots, seed=1234):
        super().__init__()
        self.nqubits = nqubits
        self.embedding = embedding
        self.shots = shots
        self.device = qml.device(backend.value, wires=nqubits)
        self.circuit_factory = CircuitFactory(nqubits, nlayers=0)
        self.kernel_circ = self._build_kernel()
        self.qkernel = None
        self.X_train = None

        # Slower
        # self.batch_kernel_circ = partial(qml.batch_input, argnum=0)(self.kernel_circ)
        # self.batch_kernel_circ = partial(qml.batch_input, argnum=1)(self.batch_kernel_circ)

        # Faster batching
        self.batch_kernel_circ = partial(qml.batch_input, argnum=1)(self.kernel_circ)

    def _build_kernel(self):
        """Build the quantum kernel using a given embedding and ansatz."""
        # Get the embedding circuit from the circuit factory
        embedding_circuit = self.circuit_factory.GetEmbeddingCircuit(self.embedding)
        adj_embedding_circuit = qml.adjoint(embedding_circuit)
        
        # Define the kernel circuit with adjoint embedding for the quantum kernel
        @qml.qnode(self.device, diff_method=None)
        def kernel(x1, x2):

            embedding_circuit(x1, wires=range(self.nqubits))
            adj_embedding_circuit(x2, wires=range(self.nqubits))

            return qml.probs(wires = range(self.nqubits))
        
        return kernel
    
    # Not used at the moment, We might be interested in computing our own kernel.
    def _quantum_kernel(self, X1, X2):
        """Calculate the quantum kernel matrix for SVM."""
        # res = np.ones((len(X1), len(X2)))
        # for i, x1 in enumerate(X1):
        #     for j, x2 in enumerate(X2):
        #         if np.array_equal(x1, x2):
        #             continue
        #         res[i, j] = self.kernel_circ(x1, x2)[0]

        # return res
        return np.array([[self.kernel_circ(x1, x2) for x2 in X2]for x1 in X1])[..., 0]
        # return np.array([self.batch_kernel_circ(x1, X2) for x1 in X1])[..., 0]
        # return np.array(self.batch_kernel_circ(X1, X2))[..., 0]

    def fit(self, X, y):
        self.X_train = X

        printer.print("\t\tTraining the SVM...")
        # self.qkernel = self._quantum_kernel(X, X)
        self.qkernel = qml.kernels.kernel_matrix(X, X, lambda x1, x2: self.kernel_circ(x1, x2)[0])

        # Train the classical SVM with the quantum kernel
        self.svm = SVC(kernel="precomputed")
        self.svm.fit(self.qkernel, y)

        printer.print("\t\tSVM training complete.")

    def predict(self, X):
        try:
            if self.X_train is None:
                raise ValueError("Model has not been fitted. Call fit() before predict().")
            
            printer.print(f"\t\t\tComputing kernel between test and training data...")
            
            # Compute kernel between test data and training data
            # kernel_test = self._quantum_kernel(X, self.X_train)
            kernel_test = qml.kernels.kernel_matrix(X, self.X_train, lambda x1, x2: self.kernel_circ(x1, x2)[0])

            if kernel_test.shape[1] == 0:
                raise ValueError(f"Invalid kernel matrix shape: {kernel_test.shape}")
            
            preds = self.svm.predict(kernel_test)
            return preds
        
        except Exception as e:
            printer.print(f"Error during prediction: {str(e)}")
            raise

    @property
    def n_params(self):
        return None


# Importing Enums
from lazyqml.Global.globalEnums import Ansatzs, Embedding
# Importing Circuits
from lazyqml.Circuits.Ansatzs import HardwareEfficient, HCzRx, TreeTensor, TwoLocal, Annular
from lazyqml.Circuits.Embeddings import DenseAngleEmbedding, ZZEmbedding, HigherOrderEmbedding

from functools import partial
import pennylane as qml

class CircuitFactory:
    def __init__(self, nqubits, nlayers) -> None:
        self.nqubits = nqubits 
        self.nlayers = nlayers

    def GetAnsatzCircuit(self,ansatz):
        if ansatz == Ansatzs.HARDWARE_EFFICIENT:
            return HardwareEfficient(self.nqubits, self.nlayers)
        elif ansatz == Ansatzs.HCZRX:
            return HCzRx(self.nqubits, self.nlayers)
        elif ansatz == Ansatzs.TREE_TENSOR:
            return TreeTensor(self.nqubits, nlayers=self.nlayers)
        elif ansatz == Ansatzs.TWO_LOCAL:
            return TwoLocal(self.nqubits, nlayers=self.nlayers)
        elif ansatz == Ansatzs.ANNULAR:
            return Annular(self.nqubits, nlayers=self.nlayers)

    def GetEmbeddingCircuit(self, embedding):
        if embedding == Embedding.RX:
            return partial(qml.AngleEmbedding, rotation='X')
        
        elif embedding == Embedding.RY:
            return partial(qml.AngleEmbedding, rotation='Y')
        
        elif embedding == Embedding.RZ:
            def RZ(inputs, wires):
                [qml.Hadamard(i) for i in wires]
                qml.AngleEmbedding(inputs, wires, rotation='Z')

            return RZ
        
        elif embedding == Embedding.ZZ:
            return ZZEmbedding
        
        elif embedding == Embedding.AMP:
            return partial(qml.AmplitudeEmbedding, pad_with=0, normalize=True)
        
        elif embedding == Embedding.DENSE_ANGLE:
            return DenseAngleEmbedding
        
        elif embedding == Embedding.HIGHER_ORDER:
            return HigherOrderEmbedding


import pennylane as qml
import torch
from time import time
from abc import abstractmethod

from lazyqml.Utils import printer, get_embedding_expressivity, find_output_shape
from lazyqml.Global.globalEnums import Backend
from lazyqml.Factories import CircuitFactory

class BaseHybridQNNModel:
    __acceptable_keys_list = ['nqubits', 'ansatz', 'embedding', 'n_class', 'layers', 'epochs', 'shots', 'lr', 'batch_size', 'device', 'backend', 'diff_method', 'seed']

    def __init__(self, **kwargs):
        for key in self.__acceptable_keys_list:
            self.__setattr__(key, kwargs.get(key))

    @property
    def n_params(self):
        return self.n_q_params + self.n_c_params

    def _build_circuit(self):
        # Get the ansatz and embedding circuits from the factory
        circuit_factory = CircuitFactory(self.nqubits, nlayers=self.layers)

        ansatz = circuit_factory.GetAnsatzCircuit(self.ansatz)
        embedding = circuit_factory.GetEmbeddingCircuit(self.embedding)

        ansatz_circ = ansatz.getCircuit()

        # Define the quantum circuit as a PennyLane qnode
        @qml.qnode(self.device, interface='torch', diff_method=self.diff_method)
        def circuit(inputs, theta):
            
            embedding(inputs, range(self.nqubits))
            ansatz_circ(theta, range(self.nqubits))

            if self.n_class==2:
                return qml.expval(qml.PauliZ(0))
            else:
                return [qml.expval(qml.PauliZ(i)) for i in range(self.n_class)]

        self.qnn = circuit
        self.n_q_params = ansatz.n_total_params

    @abstractmethod
    def _build_model(self, X):
        raise NotImplementedError()

    def fit(self, X, y):
        # Construct model
        self._build_model(X)

        # Select loss function
        if self.n_class==2:
            self.criterion = torch.nn.BCEWithLogitsLoss()
        else:
            self.criterion = torch.nn.CrossEntropyLoss()

        # Move the model to the appropriate device (GPU or CPU)
        self.device = torch.device("cuda:0" if torch.cuda.is_available() and self.device == Backend.lightningGPU else "cpu")

        # Convert training data to torch tensors and transfer to device
        X_train = torch.tensor(X, dtype=torch.float32, requires_grad=False).to(self.device)
        if self.n_class == 2:
            y_train = torch.tensor(y, dtype=torch.float32).to(self.device)
        else:
            y_train = torch.tensor(y, dtype=torch.long).to(self.device)

        # Define optimizer
        self.opt = torch.optim.Adam(self.model.parameters(), lr=self.lr)

        # Create data loader for batching
        data_loader = torch.utils.data.DataLoader(
            list(zip(X_train, y_train)), batch_size=self.batch_size, shuffle=True, drop_last=True
        )

        start_time = time()

        for epoch in range(self.epochs):
            epoch_loss = 0.0
            for i, (batch_X, batch_y) in enumerate(data_loader):
                self.opt.zero_grad()

                # Forward pass
                predictions = self.model(batch_X)
                
                # Compute loss
                loss = self.criterion(predictions, batch_y)
                loss.backward()

                # Optimization step
                self.opt.step()
                epoch_loss += loss.item()

            # Print the average loss for the epoch
            printer.print(f"\t\tEpoch {epoch+1}/{self.epochs}, Loss: {epoch_loss/len(data_loader):.4f}")

        printer.print(f"\t\tTraining completed in {time() - start_time:.2f} seconds")

        return self

    def predict(self, X):
        # Convert test data to torch tensors
        X_test = torch.tensor(X, dtype=torch.float32).to(self.device)

        # Forward pass for prediction
        y_pred = self.model(X_test)
        
        if self.n_class == 2:
            # For binary classification, apply sigmoid to get probabilities
            y_pred = torch.sigmoid(y_pred.view(-1))  # Ensure shape is [batch_size]
            # Return class labels based on a 0.5 threshold
            return (y_pred > 0.5).cpu().detach().numpy()  # Returns 0 or 1
        else:
            # For multi-class classification, y_pred is logits of shape [batch_size, n_class]
            # Return the class with the highest logit value
            return torch.argmax(y_pred, dim=1).cpu().detach().numpy()  # Returns class indices

class BasicHybridModel(BaseHybridQNNModel):
    def _build_model(self, _):
        # Builds the quantum circuit of the model
        self._build_circuit()

        # Get number of features needed for dataflow between classic layer and quantum circuit 
        in_features_qcircuit = get_embedding_expressivity(self.nqubits, self.embedding)

        # Setup of the layers of the (torch) sequential model
        weight_shapes = {"theta": self.n_q_params}
        self.qlayer = qml.qnn.TorchLayer(self.qnn, weight_shapes)
        self.clayer = torch.nn.Linear(self.nqubits, in_features_qcircuit)

        # Calculate number of classical parameters
        self.n_c_params = sum(p.numel() for p in self.clayer.parameters())

        # Constructs the final torch model
        self.model_layers = [self.clayer, self.qlayer]
        self.model = torch.nn.Sequential(*self.model_layers)


# Importing from
from lazyqml.Global.globalEnums import *
from lazyqml.Preprocessing import PCAHelper, Sanitizer

class PreprocessingFactory:
    def __init__(self, nqubits) -> None:
        self.nqubits = nqubits

    def GetSanitizer(self, imputerCat, imputerNum):
        return Sanitizer(imputerCat, imputerNum)

    def GetPreprocessing(self, embedding, ansatz):
        if embedding == Embedding.AMP and ansatz == Ansatzs.TREE_TENSOR:
            return PCAHelper(self.nqubits, 2**(2**(self.nqubits.bit_length()-1)))
        elif embedding == Embedding.AMP:
            return PCAHelper(self.nqubits, 2**self.nqubits)
        elif embedding == Embedding.DENSE_ANGLE:
            return PCAHelper(self.nqubits, 2*self.nqubits)
        elif ansatz == Ansatzs.TREE_TENSOR:
            return PCAHelper(self.nqubits, 2**(self.nqubits.bit_length()-1))
        else:
            return PCAHelper(self.nqubits, self.nqubits)


# Importing from
    # Internal Dependencies
from lazyqml.Factories import ModelFactory, PreprocessingFactory
from lazyqml.Global.globalEnums import Model, Backend
from .Tasks import QMLTask
from lazyqml.Utils import printer, calculate_free_memory, get_simulation_type, calculate_free_video_memory, generate_cv_indices, create_combinations, calculate_quantum_memory, get_train_test_split, dataProcessing

    # External Libraries
import numpy as np
import pandas as pd
import psutil
from threadpoolctl   import threadpool_limits
from sklearn.metrics import f1_score, accuracy_score, balanced_accuracy_score
from multiprocessing import Queue, Process, Pool, Manager
import queue
import time
from time import sleep

class Dispatcher:
    def __init__(self,
                nqubits,
                randomstate,
                predictions,
                shots,
                numPredictors,
                numLayers,
                classifiers,
                ansatzs,
                # backend,
                embeddings,
                numFeatures,
                learningRate,
                epochs,
                runs,
                batch,
                numSamples,
                customMetric,
                customImputerNum,
                customImputerCat,
                sequential=False,
                threshold=14,
                time=True,
                cores=-1
                ):

        self.sequential = sequential
        self.threshold = threshold
        self.timeM = time
        self.cores = cores

        self.nqubits = nqubits
        self.randomstate = randomstate
        self.shots = shots
        self.numPredictors = numPredictors
        self.numLayers = numLayers
        self.classifiers = classifiers
        self.ansatzs = ansatzs
        # self.backend = backend
        self.embeddings = embeddings
        self.learningRate = learningRate
        self.epochs = epochs
        self.batch = batch
        self.numSamples = numSamples
        self.numFeatures = numFeatures
        self.customMetric = customMetric
        self.customImputerNum = customImputerNum
        self.customImputerCat = customImputerCat
        self.predictions = predictions

    def execute_model(self, id, model_params, X_train, y_train, X_test, y_test, customMetric):
        model = ModelFactory().getModel(**model_params)
        accuracy, b_accuracy, f1, custom = 0, 0, 0, 0

        start = time.perf_counter()
        # To force limit the ammount of threads the experiment uses
        if model_params["model"] == Model.QNN:
            with threadpool_limits(limits=1):
                model.fit(X=X_train, y=y_train)
                y_pred = model.predict(X=X_test)
        else:
            model.fit(X=X_train, y=y_train)
            y_pred = model.predict(X=X_test)
        exeT = time.perf_counter() - start

        accuracy += accuracy_score(y_test, y_pred, normalize=True)
        b_accuracy += balanced_accuracy_score(y_test, y_pred)
        f1 = f1_score(y_test, y_pred, average="weighted")
        if customMetric is not None:
            custom = customMetric(y_test, y_pred)

        # Construct dataframe with results
        dict_keys = ['nqubits', 'model', 'embedding', 'ansatz', 'n_features', 'n_samples']
        model_attr = {key: model_params[key] for key in dict_keys}

        metric_results = {
            "Time taken": exeT,
            "Accuracy": accuracy,
            "Balanced Accuracy": b_accuracy,
            "F1 Score": f1,
            "Custom Metric": custom,
            "Predictions": 0
        }

        result = pd.DataFrame([{'id': id, **model_attr, 'features': model.n_params, **metric_results}])

        return result
    
    def _print_exception(self, e: Exception):
        printer.print(f"Error in the batch: {str(e)}")

    def process_gpu_task(self, gpu_queue, results):
        while not gpu_queue.empty():
            try:
                qmltask = gpu_queue.get_nowait()

                partial_result = self.execute_model(*qmltask.get_task_params())

                # Store results
                results.append(partial_result)

            except queue.Empty:
                break
            except Exception as e:
                self._print_exception(e)

    def process_cpu_task(self, cpu_queue, gpu_queue, results):
        numProcs = psutil.cpu_count(logical=False)
        total_memory = calculate_free_memory()
        available_memory = total_memory
        if not self.sequential:
            if self.cores == -1:
                available_cores = numProcs
            else:
                available_cores = self.cores
                # print(f"Available cores: {available_cores}")
        else:
            available_cores = 1

        # Lock para el acceso seguro a los recursos compartidos
        manager = Manager()
        resource_lock = manager.Lock()

        while not cpu_queue.empty():
            try:
                # Determinar número de cores a usar basado en el estado de gpu_queue
                if gpu_queue.empty():
                    max_cores = numProcs
                else:
                    max_cores = max(1, numProcs - 1)

                current_batch = []
                current_cores = 0

                # Recolectar items para procesar mientras haya recursos disponibles
                while current_cores < max_cores and not cpu_queue.empty():
                    try:
                        qmltask = cpu_queue.get_nowait()
                        # printer.print(f"ITEM CPU: {item[0]}")
                        mem_model = qmltask.model_memory

                        # Verificar si hay recursos suficientes
                        with resource_lock:
                            if available_memory >= mem_model and available_cores >= 1:
                                # printer.print(f"Available Resources - Memory: {available_memory}, Cores: {available_cores}")
                                available_memory -= mem_model
                                available_cores -= 1
                                current_batch.append(qmltask)
                                current_cores += 1
                            else:
                                # printer.print(f"Unavailable Resources - Requirements: {mem_model}, Available: {available_memory}")
                                cpu_queue.put(qmltask)
                                break

                    except queue.Empty:
                        break

                # Procesar el batch actual si no está vacío
                if current_batch:
                    # printer.print(f"Executing Batch of {len(current_batch)} Jobs")
                    with Pool(processes=len(current_batch)) as pool:
                        # Usamos map de forma síncrona para asegurar que todos los items se procesen
                        batch_results = pool.starmap(self.execute_model, [qmltask.get_task_params() for qmltask in current_batch])

                        # Filtramos los resultados None (errores) y los añadimos a results
                        # valid_results = [r for r in batch_results if r is not None]
                        results.extend(batch_results)

                    # Liberar recursos después del procesamiento
                    with resource_lock:
                        # printer.print("Freeing Up Resources")
                        for qmltask in current_batch:
                            mem_model = qmltask.model_memory
                            available_memory += mem_model
                            available_cores += 1
                            # printer.print(f"Freed - Memory: {available_memory}MB, Cores: {available_cores}")

                # printer.print("Waiting for next batch")
                sleep(0.1)

            except Exception as e:
                self._print_exception(e)
                import traceback
                traceback.print_exc()
                break

    def dispatch(self, X, y, showTable, folds=10, repeats=5, mode="cross-validation", testsize=0.3):
        """
        ################################################################################
        Preparing Data Structures & Initializing Variables
        ################################################################################
        """
        # Replace the list-based queues with multiprocessing queues
        manager = Manager()
        gpu_queue = Queue()
        cpu_queue = Queue()
        # Shared list for results if needed
        results = manager.list()
        # Also keep track of items for printing
        cpu_items = []
        gpu_items = []

        tensor_sim = get_simulation_type() == "tensor"

        RAM = calculate_free_memory()
        VRAM = calculate_free_video_memory()

        """
        ################################################################################
        Generate CV indices
        ################################################################################
        """
        cv_indices = generate_cv_indices(
            X, y,
            mode=mode,
            n_splits=folds,
            n_repeats=repeats,
            random_state=self.randomstate,
            test_size=testsize
        )

        """
        ################################################################################
        Generating Combinations
        ################################################################################
        """

        t_pre = time.perf_counter()
        combinations = create_combinations(qubits=self.nqubits,
                                        classifiers=self.classifiers,
                                        embeddings=self.embeddings,
                                        features=self.numFeatures,
                                        ansatzs=self.ansatzs,
                                        repeats=repeats,
                                        folds=folds)
        cancelledQubits = set()
        to_remove = []

        # print(combinations)

        for _, combination in enumerate(combinations):
            modelMem = combination[-1]
            if modelMem > RAM and modelMem > VRAM:
                to_remove.append(combination)

        for combination in to_remove:
            combinations.remove(combination)
            cancelledQubits.add(combination[0])

        for val in cancelledQubits:
            printer.print(f"Execution with {val} Qubits are cancelled due to memory constrains -> Memory Required: {calculate_quantum_memory(val)/1024:.2f}GB Out of {calculate_free_memory()/1024:.2f}GB")

        X = pd.DataFrame(X)

        # Prepare all model executions
        for combination in combinations:
            id, qubits, name, embedding, ansatz, n_features, repeat, fold, memModel = combination
            # feature = feature if feature is not None else "~"

            # Get indices for this repeat/fold combination
            train_idx, test_idx = get_train_test_split(cv_indices, repeat, fold)

            n_classes = len(np.unique(y))
            adjustedQubits = qubits  # or use adjustQubits if needed
            prepFactory = PreprocessingFactory(adjustedQubits)

            # Process data for this specific combination using pre-generated indices
            X_train_processed, X_test_processed, y_train_processed, y_test_processed = dataProcessing(
                X,
                y,
                prepFactory,
                self.customImputerCat,
                self.customImputerNum,
                train_idx,
                test_idx,
                ansatz=ansatz,
                embedding=embedding
            )

            # X_train, X_test, Y_train, Y_test = train_test_split(X, y, test_size=0.3, stratify=y, random_state=0)
            # print()
            # print(X_train[:5], X_test[:5])

            model_factory_params = {
                "nqubits": adjustedQubits,
                "model": name,
                "embedding": embedding,
                "ansatz": ansatz,
                "n_class": n_classes,
                "shots": self.shots,
                "seed": self.randomstate*repeat,
                "layers": self.numLayers,
                "n_samples": self.numSamples,
                "n_features": n_features,
                "lr": self.learningRate,
                "batch_size": self.batch,
                "epochs": self.epochs,
                "numPredictors": self.numPredictors
            }

            qmltask = QMLTask(
                id=id,
                model_memory=memModel,
                X_train=X_train_processed,
                X_test=X_test_processed,
                y_train=y_train_processed,
                y_test=y_test_processed,
                custom_metric=self.customMetric
            )

            # When adding items to queues
            if name == Model.QNN and qubits >= self.threshold and VRAM > memModel:
                model_factory_params["backend"] = Backend.lightningGPU if not tensor_sim else Backend.lightningTensor

                qmltask.model_params = model_factory_params
                gpu_queue.put(qmltask)

                gpu_items.append(combination)
            else:
                model_factory_params["backend"] = Backend.lightningQubit if not tensor_sim else Backend.defaultTensor

                qmltask.model_params = model_factory_params
                cpu_queue.put(qmltask)

                cpu_items.append(combination)

        if self.timeM:
            printer.print(f"PREPROCESSING TIME: {time.perf_counter()-t_pre}")

        """
        ################################################################################
        Creating processes
        ################################################################################
        """
        # Wait a bit to add remaining tasks to queue
        sleep(0.1)

        executionTime = time.perf_counter()
        gpu_process = None
        # Start GPU process
        if not gpu_queue.empty():
            gpu_process = Process(target=self.process_gpu_task, args=(gpu_queue, results))
            gpu_process.start()

        # Start CPU processes
        if not cpu_queue.empty():
            self.process_cpu_task(cpu_queue, gpu_queue, results)

        # Wait for all processes to complete
        if gpu_process is not None:
            gpu_process.join()

        executionTime = time.perf_counter()-executionTime
        printer.print(f"Execution TIME: {executionTime}")

        """
        ################################################################################
        Processing results
        ################################################################################
        """
        t_res = time.perf_counter()

        all_results = pd.concat(list(results)).reset_index(drop=True)

        scores = all_results.groupby(['id']).agg({
            'nqubits': 'first',
            'model': 'first',
            'embedding': 'first',
            'ansatz': 'first',
            'features': 'first',
            'n_features': 'first',
            'n_samples': 'first',
            'Time taken': 'sum',
            'Accuracy': 'mean',
            'Balanced Accuracy': 'mean',
            'F1 Score': 'mean',
            'Custom Metric': 'mean'
        })

        scores.loc[scores['model'] != Model.QNN_BAG, 'n_samples'] = np.nan
 
        # Clean and format dataframe
        if not self.customMetric:
            scores = scores.drop(columns=['Custom Metric'])

        scores.columns = ["Qubits", "Model", "Embedding", "Ansatz", "Features", "% Features", "% Samples", "Time taken", "Accuracy", "Balanced Accuracy", "F1 Score"]

        # Remove columns if all empty
        scores = scores.dropna(how='all', axis=1)

        # Fill values if empty
        scores = scores.fillna("~")

        # Sort scores
        scores = scores.sort_values(by="Balanced Accuracy", ascending=False).reset_index(drop=True)

        if showTable:
            print(scores.to_markdown())

        if self.timeM:
            printer.print(f"RESULTS TIME: {time.perf_counter() - t_res}")

        return scores



from dataclasses import dataclass
from typing import Union

from lazyqml.Global import Model, Embedding, Ansatzs

@dataclass
class QMLTask:
    id: int
    model_memory: float
    
    X_train: any
    X_test: any
    y_train: any
    y_test: any

    custom_metric: any
    model_params: dict = None

    def get_model_params(self):
        pass

    def get_data(self):
        return (self.X_train, self.y_train, self.X_test, self.y_test)
    
    def get_task_params(self):
        # print((self.model_params, *self.get_data(), self.custom_metric))
        return (self.id, self.model_params, *self.get_data(), self.custom_metric)


from lazyqml.Interfaces.iAnsatz import Ansatz
import pennylane as qml

class Annular(Ansatz):
    def getCircuit(self):
        def annular(theta, wires):
            """Implements an annular ansatz circuit.

            Args:
                theta (array[float]): array of parameters for the ansatz circuit
                wires (Sequence[int]): wires that the ansatz circuit acts on

            Returns:
                None
            """

            N=len(wires)

            param_count = 0

            for _ in range(self.nlayers):
                for i in range(N):
                    qml.X(wires=i)
                    qml.Hadamard(wires=i)

                for i in range(N - 1):
                    qml.CNOT(wires = [i, i + 1])
                    qml.RY(theta[param_count], wires = i+1)

                    param_count += 1

                qml.CNOT(wires=[N-1, 0])
                qml.RY(theta[param_count], wires = 0)
                param_count += 1

        return annular
    
    @property
    def n_ansatz_params(self):
        return self.nqubits


# Importing from
from lazyqml.Interfaces.iPreprocessing import Preprocessing
from sklearn.decomposition import PCA

class PCAHelper(Preprocessing):
    def __init__(self, nqubits, ncomponents):
        self.nqubits = nqubits
        self.ncomponents = ncomponents
        self.fitted = False
        self.preprocessing = PCA(n_components=self.ncomponents)

    def fit(self, X, y):
        if self.ncomponents <= X.shape[1]:
            self.fitted = True
            fitted_X = self.preprocessing.fit(X, y)
        else:
            fitted_X = X
        return fitted_X

    def fit_transform(self, X, y):
        if self.ncomponents <= X.shape[1]:
            self.fitted = True
            fitted_X = self.preprocessing.fit_transform(X, y)
        else:
            fitted_X = X
        return fitted_X

    def transform(self, X):
        if self.fitted:
            fitted_X = self.preprocessing.transform(X)
        else:
            fitted_X = X
        return fitted_X



# from .QNNTorch import QNNTorch
from .QNNBag import QNNBag
from .FastQSVM import QSVM
from .FastQNN import QNNTorch
from .QKNN import QKNN
from .BaseHybridModel import BaseHybridQNNModel, BasicHybridModel

__all__ = ['QNNTorch', 'QNNBag', 'QSVM', 'QKNN', 'BaseHybridQNNModel', 'BasicHybridModel']


import pennylane as qml
from pennylane.operation import Operation, AnyWires

class HigherOrderEmbedding(Operation):
    num_wires = AnyWires
    grad_method = None

    def __init__(self, features, wires, id=None):

        shape = qml.math.shape(features)[-1:]
        n_features = shape[0]
        if n_features > len(wires):
            raise ValueError(
                f"Features must be of length {len(wires)} or less; got length {n_features}."
            )

        self._hyperparameters = {}

        wires = wires[:n_features]
        super().__init__(features, wires=wires, id=id)


    @property
    def num_params(self):
        return 1

    @staticmethod
    def compute_decomposition(features, wires):

        batched = qml.math.ndim(features) > 1
        features = qml.math.T(features) if batched else features

        op_list = []

        op_list.extend([qml.RY(features[i], i) for i in wires])

        for i in wires[1:]:
            op_list.append(qml.CNOT(wires = [i - 1, i]))
            op_list.append(qml.RY(features[i - 1]*features[i], wires=i))

        return op_list


from lazyqml.Interfaces.iAnsatz import Ansatz
import pennylane as qml

class HCzRx(Ansatz):
    def getCircuit(self):
        def HCzRx(theta, wires):
            """Implements an ansatz circuit composed of Hadamard, CZ, and RX gates.

            Args:
                theta (array[float]): array of parameters for the ansatz circuit
                wires (Sequence[int]): wires that the ansatz circuit acts on

            Returns:
                None
            """
            N = len(wires)

            param_count = 0

            for _ in range(int(self.nlayers)):
                for i in range(N):
                    qml.Hadamard(wires = wires[i])
                
                for i in range(N-1):
                    qml.CZ(wires=[wires[i], wires[i+1]])
                qml.CZ(wires=[wires[N-1],wires[0]])
                
                for i in range(N):
                    qml.RX(theta[param_count], wires=wires[i])
                    param_count += 1

        return HCzRx
    
    @property
    def n_ansatz_params(self):
        return self.nqubits


import sys
import gc
import numpy     as np
import pennylane as qml
import psutil
import warnings

from contextlib              import nullcontext
from sklearn.svm             import SVC
from threadpoolctl           import threadpool_limits

import numpy as np
from sklearn.svm import SVC
import pennylane as qml
from lazyqml.Factories import CircuitFactory
from lazyqml.Utils import printer, _numpy_math_api
from lazyqml.Interfaces.iModel import Model

from functools import partial

from time import time

# import torch
# from itertools import product
# from itertools import combinations_with_replacement

from threadpoolctl import threadpool_limits

# from sys import getsizeof

# -----------------------------------------------------------------------------
# Parámetros globales de memoria
# -----------------------------------------------------------------------------

##### Esto va en Global._config
# Faltaria añadir funciones para obtener estos valores en Utils

MEM_Threshold = 0.75
MEM_Safety    = 1.60
MEM_GBytes    = float(psutil.virtual_memory().available) / (1024**3)

STATE_DTYPE   = np.complex128   # dtype de los statevectors
KERNEL_DTYPE  = np.float64      # dtype de la matriz kernel (real)

warnings.filterwarnings("ignore")

# -----------------------------------------------------------------------------
# QSVM Fast_Block
# -----------------------------------------------------------------------------

class QSVM(Model):
    def __init__(self,
            nqubits,
            embedding,
            qnode,
            cores=1,
            max_cached_states=None,
            ram_gb=MEM_GBytes,
            fraction=MEM_Threshold,
            safety=MEM_Safety,
            state_dtype=STATE_DTYPE,
            kernel_dtype=KERNEL_DTYPE
        ):
        super().__init__()

        # self.nqubits = nqubits
        # self.embedding = embedding
        # self.shots = shots
        # self.device = qml.device(backend.value, wires=nqubits)
        # self.device = qml.device('default.qubit', wires=nqubits)
        self.circuit_factory = CircuitFactory(nqubits, nlayers=0)
        # self.kernel_circ = self._build_kernel()
        # self.qkernel = None
        # self.X_train = None

        self.nqubits           = nqubits
        self.qnode             = qnode
        self.embedding_circuit = self.circuit_factory.GetEmbeddingCircuit(embedding)
        self.cores             = cores
        self.kernel_circ       = self._build_kernel()
        self.qkernel           = None
        self.X_train           = None

        self.state_dtype       = state_dtype
        self.kernel_dtype      = kernel_dtype
        self.ram_gb            = ram_gb
        self.ram_fraction      = fraction
        self.safety            = safety

        self.numpy_api = _numpy_math_api()

        if max_cached_states is None:
            self.max_cached_states = self._estimate_max_cached_states()
        else:
            self.max_cached_states = max_cached_states

        # Slower
        # self.batch_kernel_circ = partial(qml.batch_input, argnum=0)(self.kernel_circ)
        # self.batch_kernel_circ = partial(qml.batch_input, argnum=1)(self.batch_kernel_circ)

        # Faster batching
        # self.batch_kernel_circ = partial(qml.batch_input, argnum=1)(self.kernel_circ)

    # -------------------------------------------------------------------------
    # max number of statevectors we can put in RAM
    # -------------------------------------------------------------------------
    def _estimate_max_cached_states(self):
        bytes_per_complex = np.dtype(self.state_dtype).itemsize
        bytes_vstate      = bytes_per_complex * (1 << self.nqubits)
        max_bytes         = self.ram_gb * (1024**3) * self.ram_fraction
        bytes_effect      = self.safety * bytes_vstate
        return max(1, int(max_bytes // bytes_effect))

    # def _build_kernel(self):
    #     """Build the quantum kernel using a given embedding and ansatz."""
    #     # Get the embedding circuit from the circuit factory
    #     # embedding_circuit = self.circuit_factory.GetEmbeddingCircuit(self.embedding)
    #     # adj_embedding_circuit = qml.adjoint(embedding_circuit)
        
    #     # # Define the kernel circuit with adjoint embedding for the quantum kernel
    #     # @qml.qnode(self.device, diff_method=None)
    #     # def kernel(x1, x2):

    #     #     embedding_circuit(x1, wires=range(self.nqubits))
    #     #     adj_embedding_circuit(x2, wires=range(self.nqubits))

    #     #     return qml.probs(wires = range(self.nqubits))
        
    #     # return kernel

    # -------------------------------------------------------------------------
    # QNode returns statevector
    # -------------------------------------------------------------------------
    def _build_kernel(self):
        """Build the quantum kernel using a given embedding and ansatz."""
        @self.qnode
        def kernel(x):
            self.embedding_circuit(x, wires=range(self.nqubits))
            return qml.state()
        return kernel

    # -------------------------------------------------------------------------
    # Context manager for BLAS threads
    # -------------------------------------------------------------------------
    def _threadpool_ctx(self):
        api = _numpy_math_api()
        if api is None:
            return nullcontext()
        return threadpool_limits(limits=self.cores, user_api=api)

    # -------------------------------------------------------------------------
    # block sizes for X1 (test) y X2 (train)
    # -------------------------------------------------------------------------
    def _compute_block_sizes(self, n1, n2):
        max_states = max(2, self.max_cached_states)
        ratio = n1 / (n1 + n2)

        if ratio < 0.2:
            frac = 0.80   # X1 very small
        elif ratio < 0.4:
            frac = 0.60   # X1 moderate
        else:
            frac = 0.50   # X1 an X2 equals

        bs1 = int(max_states * frac)
        bs2 = max_states - bs1

        bs1 = max(1, min(n1, bs1))
        bs2 = max(1, min(n2, bs2))
        return bs1, bs2
    
    # -------------------------------------------------------------------------
    # MODO C CPU: bloques en RAM
    # -------------------------------------------------------------------------
    def _quantum_kernel_block(self, X1, X2, is_symmetric=False):
        n1       = X1.shape[0]  # nº muestras X1 (test)
        n2       = X2.shape[0]  # nº muestras X2 (train)
        bs1, bs2 = self._compute_block_sizes(n1, n2)
        dim      = 1 << self.nqubits

        kernel_matrix = np.empty((n1, n2),   dtype=self.kernel_dtype)
        x_buf         = np.empty((bs1, dim), dtype=self.state_dtype)
        y_buf         = np.empty((bs2, dim), dtype=self.state_dtype)
        k_buf         = np.empty((bs1, bs2), dtype=self.kernel_dtype)

        if is_symmetric:
            for i_start in range(0, n1, bs1):
                i_end = min(i_start + bs1, n1)
                m1    = i_end - i_start

                with threadpool_limits(limits=1, user_api=self.numpy_api):
                    for k in range(m1):
                        x_buf[k, :] = self.kernel_circ(X1[i_start + k])
                x_view = x_buf[:m1, :]

                for j_start in range(i_start, n2, bs2):
                    j_end = min(j_start + bs2, n2)
                    m2    = j_end - j_start

                    with threadpool_limits(limits=1, user_api=self.numpy_api):
                        for k in range(m2):
                            y_buf[k, :] = self.kernel_circ(X1[j_start + k])
                    y_view = y_buf[:m2, :]

                    with self._threadpool_ctx():
                        gram = x_view @ y_view.conj().T
                    k_view = k_buf[:m1, :m2]
                    np.square(gram.real, out=k_view)
                    k_view += gram.imag * gram.imag

                    kernel_matrix[i_start:i_end, j_start:j_end] = k_view
                    if j_start != i_start:
                        kernel_matrix[j_start:j_end, i_start:i_end] = k_view.T
        else:
            for i_start in range(0, n1, bs1):
                i_end = min(i_start + bs1, n1)
                m1    = i_end - i_start

                with threadpool_limits(limits=1, user_api=self.numpy_api):
                    for k in range(m1):
                        x_buf[k, :] = self.kernel_circ(X1[i_start + k])
                x_view = x_buf[:m1, :]

                for j_start in range(0, n2, bs2):
                    j_end = min(j_start + bs2, n2)
                    m2    = j_end - j_start

                    with threadpool_limits(limits=1, user_api=self.numpy_api):
                        for k in range(m2):
                            y_buf[k, :] = self.kernel_circ(X2[j_start + k])
                    y_view = y_buf[:m2, :]

                    with self._threadpool_ctx():
                        gram = x_view @ y_view.conj().T
                    k_view = k_buf[:m1, :m2]
                    np.square(gram.real, out=k_view)
                    k_view += gram.imag * gram.imag

                    kernel_matrix[i_start:i_end, j_start:j_end] = k_view
        return kernel_matrix

    # -------------------------------------------------------------------------
    # Dispatcher: for now only A and C modes in CPU
    # -------------------------------------------------------------------------
    def _quantum_kernel(self, X1, X2):
        """Calculate the quantum kernel matrix for SVM."""

        are_equal = np.array_equal(X1, X2)

        dim = 1 << self.nqubits
        n1  = X1.shape[0] # if (X1 != X2) -> X1 is X_test
        n2  = X2.shape[0] # if (X1 != X2) -> X2 is X_train

        if are_equal:
            A_mode = (n1 <= self.max_cached_states)
        else:
            A_mode = (n1 + n2 <= self.max_cached_states)

        if A_mode:
            x_sv = np.empty((n1, dim), dtype=self.state_dtype)
            if not are_equal:
                y_sv = np.empty((n2, dim), dtype=self.state_dtype)

            # We can use batching (x_sv = self.kernel_circ(X1)), but not all backend works with it
            with threadpool_limits(limits=1, user_api=self.numpy_api):
                for k in range(n1):
                    x_sv[k, :] = self.kernel_circ(X1[k])

            if are_equal:
                y_sv = x_sv
            else:
                with threadpool_limits(limits=1, user_api=self.numpy_api):
                    for k in range(n2):
                        y_sv[k, :] = self.kernel_circ(X2[k])

            with self._threadpool_ctx():
                gram = x_sv @ y_sv.conj().T

            R = np.empty(gram.shape, dtype=self.kernel_dtype)
            np.square(gram.real, out=R)
            R += gram.imag * gram.imag

            return R
        else:
            # C Mode in CPU
            return self._quantum_kernel_block(X1, X2, is_symmetric=are_equal)

    # -------------------------------------------------------------------------
    # API fit / predict
    # -------------------------------------------------------------------------
    def fit(self, X, y):
        self.X_train = X
        printer.print("\t\tTraining the SVM...")

        self.qkernel = self._quantum_kernel(X, X)
        self.svm     = SVC(kernel="precomputed")
        self.svm.fit(self.qkernel, y)
        printer.print("\t\tSVM training complete.")


    def predict(self, X):
        if self.X_train is None:
            raise ValueError("Model has not been fitted. Call fit() before predict().")
        
        printer.print(f"\t\t\tComputing kernel between test and training data...")

        kernel_test = self._quantum_kernel(X, self.X_train)

        if kernel_test.shape[1] == 0:
            raise ValueError(f"Invalid kernel matrix shape: {kernel_test.shape}")

        preds = self.svm.predict(kernel_test)
        return preds

    @property
    def n_params(self):
        return None



import torch
import pennylane as qml
from time import time
import numpy as np
import torchvision.models as torchmodels

from lazyqml.Interfaces.iModel import Model
from lazyqml.Interfaces.iAnsatz import Ansatz
from lazyqml.Interfaces.iCircuit import Circuit
from lazyqml.Factories import CircuitFactory
from lazyqml.Global.globalEnums import Backend
from lazyqml.Utils import printer, get_simulation_type, get_max_bond_dim

import warnings

# https://glassboxmedicine.com/2020/12/08/using-predefined-and-pretrained-cnns-in-pytorch-tutorial-with-code/

class QCNN(Model):
    def __init__(self, nqubits, backend, ansatz, embedding, n_class, layers, epochs, shots, lr, batch_size, seed=1234) -> None:
        super().__init__()
        self.nqubits = nqubits
        self.ansatz = ansatz
        self.shots = shots
        self.embedding = embedding
        self.n_class = n_class
        self.layers = layers
        self.epochs = epochs
        self.lr = lr
        self.batch_size = batch_size
        self.backend = backend

        if get_simulation_type() == "tensor":
            if backend != Backend.lightningTensor:
                device_kwargs = {
                    "max_bond_dim": get_max_bond_dim(),
                    "cutoff": np.finfo(np.complex128).eps,
                    "contract": "auto-mps",
                }
            else:
                device_kwargs = {
                    "max_bond_dim": get_max_bond_dim(),
                    "cutoff": 1e-10,
                    "cutoff_mode": "abs",
                }
                
            self.deviceQ = qml.device(backend.value, wires=nqubits, method='mps', **device_kwargs)
        else:
            self.deviceQ = qml.device(backend.value, wires=nqubits)

        self.device = None
        self.params_per_layer = None
        self.circuit_factory = CircuitFactory(nqubits,nlayers=layers)
        self.qnn = None
        self.params = None

        # Traditional CNN
        full_cnn = torchmodels.resnet18(pretrained=False)
        self.cnn = torch.nn.Sequential(*(list(full_cnn.children())[0:8]))

        self._build_circuit()

        # Suppress all warnings
        warnings.filterwarnings("ignore")

        # Initialize PyTorch optimizer and loss function
        self.opt = None  # Will initialize in fit method with model parameters
        if self.n_class==2:
            self.criterion = torch.nn.BCEWithLogitsLoss()
        else:
            self.criterion = torch.nn.CrossEntropyLoss()


    def _build_circuit(self):
        # Get the ansatz and embedding circuits from the factory
        ansatz: Ansatz = self.circuit_factory.GetAnsatzCircuit(self.ansatz)
        embedding: Circuit = self.circuit_factory.GetEmbeddingCircuit(self.embedding)

        # Retrieve parameters per layer from the ansatz
        self.params_per_layer = ansatz.getParameters()
        

        # Define the quantum circuit as a PennyLane qnode
        @qml.qnode(self.deviceQ, interface='torch', diff_method='adjoint' if get_simulation_type() == "statevector" else 'best')
        def circuit(x, theta):
            
            embedding.getCircuit()(x, wires=range(self.nqubits))
            
            ansatz.getCircuit()(theta, wires=range(self.nqubits))
            if self.n_class==2:
                return qml.expval(qml.PauliZ(0))
            else:
                return [qml.expval(qml.PauliZ(wires=n)) for n in range(self.n_class)]
            

        self.qnn = circuit

    def forward(self, x, theta):

        cnn_output = self.cnn


        qnn_output = self.qnn(x, theta)
        if self.n_class == 2:
            #return (qnn_output + 1) / 2
            return qnn_output.squeeze()
        else:
            # If qnn_output is a list, apply the transformation to each element
            #return torch.tensor([(output + 1) / 2 for output in qnn_output])
            return torch.stack([output for output in qnn_output]).T
        #return (self.qnn(x, theta) + 1)/2

    def fit(self, X, y):
        # Move the model to the appropriate device (GPU or CPU)
        self.device = torch.device("cuda:0" if torch.cuda.is_available() and self.backend == Backend.lightningGPU else "cpu")
        # print(f"USING: {self.device} and {self.deviceQ}")

        # Convert training data to torch tensors and transfer to device
        X_train = torch.tensor(X, dtype=torch.float32).to(self.device)
        if self.n_class == 2:
            y_train = torch.tensor(y, dtype=torch.float32).to(self.device)
        else:
            y_train = torch.tensor(y, dtype=torch.long).to(self.device)


        # Initialize parameters as torch tensors
        num_params = int(self.layers * self.params_per_layer)
        printer.print(f"\t\tInitializing {num_params} parameters")
        self.params = torch.randn((num_params,), device=self.device, requires_grad=True)  # Ensure params are on the same device

        # Define optimizer
        self.opt = torch.optim.Adam([self.params], lr=self.lr)

        # Create data loader for batching
        data_loader = torch.utils.data.DataLoader(
            list(zip(X_train, y_train)), batch_size=self.batch_size, shuffle=True, drop_last=True
        )
        
        start_time = time()

        for epoch in range(self.epochs):
            epoch_loss = 0.0
            for batch_X, batch_y in data_loader:
                self.opt.zero_grad()

                # Forward pass
                batch_X, batch_y = batch_X.to(self.device), batch_y.to(self.device)  # Ensure batch data is on the same device
                predictions = torch.stack([self.forward(x, self.params) for x in batch_X])
                # Compute loss
                loss = self.criterion(predictions, batch_y)  # Ensure all tensors are on the same device
                loss.backward()

                # Optimization step
                self.opt.step()
                epoch_loss += loss.item()

            # Print the average loss for the epoch
            printer.print(f"\t\tEpoch {epoch+1}/{self.epochs}, Loss: {epoch_loss/len(data_loader):.4f}")

        printer.print(f"\t\tTraining completed in {time() - start_time:.2f} seconds")
        self.params = self.params.detach().cpu()  # Save trained parameters to CPU

    def predict(self, X):
        # Convert test data to torch tensors
        X_test = torch.tensor(X, dtype=torch.float32).to(self.device)
        
        # Forward pass for prediction
        y_pred = torch.stack([self.forward(x, self.params) for x in X_test])
        
        if self.n_class == 2:
            # For binary classification, apply sigmoid to get probabilities
            y_pred = torch.sigmoid(y_pred.view(-1))  # Ensure shape is [batch_size]
            # Return class labels based on a 0.5 threshold
            return (y_pred > 0.5).cpu().numpy()  # Returns 0 or 1
        else:
            # For multi-class classification, y_pred is logits of shape [batch_size, n_class]
            # Return the class with the highest logit value
            return torch.argmax(y_pred, dim=1).cpu().numpy()  # Returns class indices

    def getTrainableParameters(self):
        print(self.params)


#!/usr/bin/env python

"""Tests for `lazyqml` package."""

import unittest

def import_data():
    from sklearn.datasets import load_iris

    # Load data
    data = load_iris()
    X = data.data
    y = data.target
    return X, y

class TestLazyqml(unittest.TestCase):
    """Tests for `lazyqml` package."""

    @classmethod
    def setUpClass(self):
        from lazyqml.Utils import get_simulation_type, get_max_bond_dim, set_simulation_type

    def _test_import(self):
        import lazyqml 
        # print("Imported correctly")

    def _test_simulation_strings(self):
        from lazyqml.Utils import get_simulation_type, get_max_bond_dim, set_simulation_type

        # Verify getter/setter of simulation type flag
        sim = "statevector"
        set_simulation_type(sim)
        self.assertTrue(get_simulation_type(), "statevector")

        sim = "tensor"
        set_simulation_type(sim)
        self.assertTrue(get_simulation_type(), "tensor")

        # Verify that ValueError is raised when number or different string is set
        sim = 3
        with self.assertRaises(ValueError):
            set_simulation_type(sim)

        sim = "tns"
        with self.assertRaises(ValueError):
            set_simulation_type(sim)

    def _test_bdim(self):
        import lazyqml

        # TODO: Desarrollar mas
        lazyqml.get_max_bond_dim()

    def test_basic_exec(self):
        from lazyqml import QuantumClassifier
        from lazyqml.Global import Embedding, Ansatzs, Model

        X, y = import_data()

        nqubits = {4, 8}
        embeddings = {Embedding.RX, Embedding.DENSE_ANGLE}
        ansatzs = {Ansatzs.TWO_LOCAL}
        models = {Model.QSVM, Model.QNN, Model.QKNN}
        layers = 2
        verbose = True
        sequential = False
        epochs = 10

        qc = QuantumClassifier(nqubits=nqubits, embeddings=embeddings, ansatzs=ansatzs, classifiers=models, numLayers=layers,
                            verbose=verbose, sequential=sequential, epochs=epochs)
        
        # if cores > 1: qc.repeated_cross_validation(X,y,n_splits=splits,n_repeats=repeats)
        # else: qc.fit(X, y)
        qc.fit(X, y)

    def _test_paper(self):
        from lazyqml import QuantumClassifier
        from lazyqml.Global import Model, Embedding, Ansatzs
        from sklearn.datasets import load_breast_cancer
        from sklearn.model_selection import train_test_split

        data = load_breast_cancer()
        X, y = data.data, data.target

        # X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=.25, random_state=4241)
        
        # Initialize QuantumClassifier
        classifier = QuantumClassifier(nqubits={4, 8, 16}, embeddings={Embedding.RX, Embedding.RY, Embedding.DENSE_ANGLE, Embedding.ZZ}, ansatzs={Ansatzs.TWO_LOCAL, Ansatzs.HARDWARE_EFFICIENT, Ansatzs.ANNULAR}, classifiers={Model.QNN}, verbose=True, sequential=False, threshold=20, epochs=5)
    
        # Fit and predict
        classifier.fit(X, y, .25)

    def _test_tensor(self):
        from lazyqml import QuantumClassifier
        from lazyqml.Global import Embedding, Ansatzs, Model, Backend
        from lazyqml.Utils import get_simulation_type, get_max_bond_dim, set_simulation_type

        set_simulation_type("tensor")
        assert get_simulation_type() == "tensor"

        X, y = import_data()

        qubits = 4
        nqubits = {qubits}
        embeddings = {Embedding.RX}
        ansatzs = {Ansatzs.TWO_LOCAL}
        models = {Model.QNN}
        epochs = 2
        layers = 1
        verbose = True
        sequential = False
        backend = Backend.lightningTensor

        qc = QuantumClassifier(nqubits=nqubits, embeddings=embeddings, ansatzs=ansatzs, classifiers=models, numLayers=layers, epochs=epochs, verbose=verbose, sequential=sequential, backend=backend)
        
        qc.fit(X, y)

    def _test_qknn(self):
        from lazyqml import QuantumClassifier
        from lazyqml.Global import Embedding, Ansatzs, Model

        X, y = import_data()

        qubits = 4
        nqubits = {4, 8}
        embeddings = {Embedding.RX, Embedding.DENSE_ANGLE}
        ansatzs = {Ansatzs.TWO_LOCAL}
        models = {Model.QSVM, Model.QKNN}
        layers = 2
        verbose = True
        sequential = False
        epochs = 10

        qc = QuantumClassifier(nqubits=nqubits, embeddings=embeddings, ansatzs=ansatzs, classifiers=models, numLayers=layers,
                            verbose=verbose, sequential=sequential, epochs=epochs)
        
        # if cores > 1: qc.repeated_cross_validation(X,y,n_splits=splits,n_repeats=repeats)
        # else: qc.fit(X, y)
        qc.fit(X, y)

# from lazyqml.Factories.Models import GradDescentModel
# import pennylane as qml
# class GDCustomModel(GradDescentModel):
#     # def __init__(self, kwargs):
#     #     super().__init__(**kwargs)

#     def trainable_circuit(self, x, theta):
#         qml.AngleEmbedding(x, wires=range(self.nqubits), rotation='Y')

#         param_count = 0
#         for _ in range(self.layers):
#             for i in range(self.nqubits):
#                 qml.RY(theta[param_count], wires = i)
#                 param_count += 1
#             for i in range(self.nqubits - 1):
#                 qml.CNOT(wires = [i, i + 1])

#     def getTrainableParameters(self):
#         return self.layers*self.nqubits

# class TestCustomModel(unittest.TestCase):
#     def _test_custom_basic(self):
#         from lazyqml import QuantumClassifier
#         from lazyqml.Global.globalEnums import Embedding, Ansatzs, Model, Backend
#         from lazyqml.Factories.Models.modelBlueprint import GenericModel

#         X, y = import_data()

#         qubits = 4
#         nqubits = {qubits}
#         embeddings = {Embedding.RX}
#         ansatzs = {Ansatzs.TWO_LOCAL}
#         models = {Model.QNN}
#         epochs = 10
#         layers = 5
#         verbose = True
#         sequential = False
#         backend = Backend.lightningQubit

#         # {
#         #     "name": custom_circuit_1,
#         #     "type": embedding, ansatz o modelo,
#         #     "circuit": objeto de la clase Model o Ansatz o Circuit,
#         #     "circ_params" (opcional): diccionaro de parametros
#         # },
        
#         # num_params, backend, n_class, epochs, shots, lr, batch_size

#         custom = [
#             {
#                 "name": "custom_1",
#                 "type": "model",
#                 "circuit": GDCustomModel,
#                 "circ_params": {
#                     "layers": layers + 1
#                 }
#             }
#         ]

#         qc = QuantumClassifier(nqubits=nqubits, embeddings=embeddings, ansatzs=ansatzs, classifiers=models, numLayers=layers, epochs=epochs, verbose=verbose, sequential=sequential, backend=backend, custom_circuits=custom)
        
#         qc.fit(X, y)

if __name__ == '__main__':
    unittest.main()